/* Copyright (C) 2005, 2006, 2007, 2008, 2009 Stefan Frenz
 * 
 * This file is part of SJC, the Small Java Compiler written by Stefan Frenz.
 * 
 * SJC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * SJC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with SJC. If not, see <http://www.gnu.org/licenses/>.
 */

package compbase;

import osio.TextPrinter;

/**
 * TypeRef: fully qualified definition of a type in any declaration 
 * 
 * @author S. Frenz
 * @version 091007 fixed inline array type compare
 *  version 091005 removed unneeded methods
 *  version 091001 adopted changed memory interface
 *  version 090918 added support for package private
 *  version 090218 added centralized access level check
 *  version 090207 added copyright notice
 *  version 080604 added checker for Throwable and Exception
 *  version 080119 fixed setting classdesc-required-flag for qid-arrays
 *  version 080105 adopted changed signature of Unit.isParent
 *  version 071227 added support for inner units
 *  version 071001 fixed isObjType
 *  version 070910 clarified null-compare in compareType
 *  version 070808 optimized getMinSize
 *  version 070731 optimized resolve
 *  version 070727 got is* from Expr
 *  version 070702 fixed un-special handling for normal arrays of special classes
 *  version 070628 adopted changes for "Expr extends TypeRef"
 *  version 070625 renamed S_* to T_* to conform with StdTypes
 *  version 070526 added support for inline arrays in instances
 *  version 070111 adapted change in printPos and compErr
 *  version 070106 replaced arrayInsideStruct by inlinedArraySize
 *  version 061129 moved initialized TypeRef-object to Context
 *  version 061111 added flag arrayInsideStruct, extended getPtrType to getRegType
 *  version 061030 changed detection of indirectCall
 *  version 060620 no implicit conversion of standard types
 *  version 060607 initial version
 */

public class TypeRef extends Token {
  private final static String ERR_NOTVIS = " is not visible";
  
  //internal types
  public final static int T_RES  = 0;
	public final static int T_BYTE = StdTypes.T_BYTE; //"byte"
	public final static int T_SHRT = StdTypes.T_SHRT; //"short"
	public final static int T_INT  = StdTypes.T_INT;  //"int"
	public final static int T_LONG = StdTypes.T_LONG; //"long"
	public final static int T_FLT  = StdTypes.T_FLT;  //"float"
	public final static int T_DBL  = StdTypes.T_DBL;  //"double"
	public final static int T_CHAR = StdTypes.T_CHAR; //"char"
	public final static int T_BOOL = StdTypes.T_BOOL; //"boolean"
  public final static int T_NULL = StdTypes.T_NULL; //null-type, unresolved
  public final static int T_NNPT = StdTypes.T_NNPT; //null-type, pointer
  public final static int T_NDPT = StdTypes.T_NDPT; //null-type, double sized pointer
  public final static int T_VOID = StdTypes.T_VOID;
  public final static int T_QID  = StdTypes.T_QID;
  
  //results of comparison
  public final static int C_NP = -1; //conversion not possible
  public final static int C_EQ =  0; //type equal
  public final static int C_TT =  1; //conversion possible, take this type
  public final static int C_OT =  2; //conversion possible, take type of compared expression
  
  //array specials
  public final static int A_NOINLINEARR =  0; //no inline array
  public final static int A_NOTSPEC     = -1; //size of struct-array not specified
  public final static int A_DONTCHECK   = -2; //index of struct-array must not be checked
  public final static int A_INSTINLARR  = -3; //normal instance-inline array
  
  //levels of access with minimum visibility level of destination
  public final static int L_NOTVISIBLE = 0; //destination is not visible
  public final static int L_PRIVATE    = 1; //minimum access level is private, which is fulfilled
  public final static int L_PACKPRIV   = 2; //minimum access level is package private, which is fulfilled
  public final static int L_PROTECTED  = 3; //minimum access level is protected, which is fulfilled
  public final static int L_PUBLIC     = 4; //minimum access level is public, which is fulfilled
  
  //variables required for resolving
  public int baseType, arrDim;
  public int inlinedArraySize; //required to check struct arrays, initialized to 0 for normal array
	public QualID qid; //used if type is non-standard
  public Object symHint; //required for symbolinformer
	
	public TypeRef(int fid, int il, int ic) {
		super(fid, il, ic);
	}
	
	public void printType(TextPrinter v) {
		int i;
		
		switch (baseType) {
			case T_RES: v.print("resType"); return;
			case T_VOID: v.print("void"); return;
			case T_QID: qid.printFullQID(v); break;
			default: StdTypes.printStdType(baseType, v);
		}
    if (inlinedArraySize!=0) v.print("(S)");
    for (i=0; i<arrDim; i++) v.print("[]");
	}
	
	public TypeRef copy() {
		TypeRef n;
		
		n=new TypeRef(fileID, line, col);
		n.baseType=baseType;
		n.arrDim=arrDim;
		n.qid=qid;
		return n;
	}
	
	public boolean resolve(Unit inUnit, Context ctx) {
	  if (baseType!=T_QID) return true;
    if (!qid.resolveAsUnit(inUnit, ctx)) return false;
    if (arrDim>0) {
      qid.unitDest.modifier|=Unit.MA_ACCSSD; //array of qid requires class descriptor to be generated
      if ((qid.unitDest.modifier&Unit.M_INDIR)!=0) {
        printPos(ctx, "arrays of interfaces not supported");
        return false;
      }
    }
    if ((qid.unitDest.modifier&Unit.M_PROT)!=0) {
      if (qid.unitDest.pack.packDest!=inUnit.pack.packDest) {
        printPos(ctx, "can not access protected class ");
        qid.printFullQID(ctx.out);
        ctx.out.print(" outside package");
        return false;
      }
    }
    return true;
	}
	
	public int compareType(TypeRef cmp, boolean searchParent, Context ctx) {
    boolean meNull;
    
    //check uninitialized null
    if (baseType==T_NULL) {
      if (cmp.baseType<0 || cmp.arrDim>0) return C_EQ;
      return C_NP;
    }
    if (cmp.baseType==T_NULL) {
      if (baseType<0 || arrDim>0) return C_EQ;
      return C_NP;
    }
    //check initialized null
    if ((meNull=(baseType==T_NNPT || baseType==T_NDPT)) || cmp.baseType==T_NNPT || baseType==T_NDPT) {
      return getRegType(ctx)!=cmp.getRegType(ctx) ? C_NP : meNull ? C_OT : C_TT;
    }
    //check additional array conditions
    if (arrDim!=0 || cmp.arrDim!=0) {
      if (inlinedArraySize!=A_NOINLINEARR || cmp.inlinedArraySize!=A_NOINLINEARR) return C_NP;
      if (arrDim!=cmp.arrDim) {
        if (cmp.baseType==T_QID && (cmp.qid.unitDest==ctx.langRoot.unitDest
            || cmp.qid.unitDest==ctx.rteSArray)) return C_OT;
        if (baseType==T_QID && (qid.unitDest==ctx.langRoot.unitDest
          || qid.unitDest==ctx.rteSArray)) return C_TT;
        return C_NP;
      }
    }
    //check normally
    if (baseType<=0 && baseType!=cmp.baseType) return C_NP; //will never match
	  if (baseType==T_VOID) return C_EQ; //void==void always
    //if (inlinedArraySize!=0 || cmp.inlinedArraySize!=0) return C_NP; //inlined array pointer not existent
	  if (baseType!=T_QID) { //check standard-conversion
	    if (baseType==cmp.baseType) return C_EQ; //same standard-type
	    if (!searchParent) return C_NP; //do not try to convert
	    return compareStdTypes(baseType, cmp.baseType);
	  }
	  //check object-conversion
	  if (qid.unitDest==cmp.qid.unitDest) return C_EQ;
	  if (!searchParent) return C_NP;
	  /*if (qid.unitDest==null || cmp.qid.unitDest==null) {
	    System.out.println("### error in qid-resolving for one of the following variables ###");
	    qid.printPos(ctx, "this");
	    System.out.println((qid.unitDest==null?"==null":" valid")+" at line "+qid.line);
	    cmp.qid.printPos(ctx, "cmp");
	    System.out.println((cmp.qid.unitDest==null?"==null":" valid")+" at line "+cmp.qid.line);
	    throw new NullPointerException();
	  }*/
	  if (qid.unitDest.isParent(cmp.qid.unitDest, ctx)) return C_TT;
	  if (cmp.qid.unitDest.isParent(qid.unitDest, ctx)) return C_OT;
	  return C_NP;
	}
  
  public int getRegType(Context ctx) {
    //reserved or void? -> no pointer
    if (baseType==T_RES || baseType==T_VOID) return 0;
    //array? -> normal pointer
    if (arrDim>0) return StdTypes.T_PTR;
    //standard-type? -> pointer only if resolved null-type
    if (baseType!=T_QID) {
      if (baseType==T_NNPT) return StdTypes.T_PTR;
      if (baseType==T_NDPT) return StdTypes.T_DPTR;
      if (baseType==T_NULL) {
        compErr(ctx, "unresolved null-type in TypeRef.getRegType");
        return 0;
      }
      return baseType;
    }
    //get size of QID-type
    if (qid.unitDest==null) {
      compErr(ctx, "unresolved type in TypeRef.getRegType");
      return 0;
    }
    if ((qid.unitDest.modifier&Unit.M_INDIR)!=0) return StdTypes.T_DPTR;
    return StdTypes.T_PTR;
  }
  
  public void getTypeOf(TypeRef from) {
    if (from==null) baseType=T_VOID; //void for constructors
    else {
      baseType=from.baseType;
      qid=from.qid;
      arrDim=from.arrDim;
      inlinedArraySize=from.inlinedArraySize;
    }
  }
  
	public static int compareStdTypes(int t1, int t2) {
	  if (t1==t2) return C_EQ;
    /*switch (t1) {
      case S_BYTE:
        if (t2==S_SHRT) return C_OT;
        if (t2==S_INT) return C_OT;
        if (t2==S_LONG) return C_OT;
        return C_NP;
      case S_SHRT:
        if (t2==S_BYTE) return C_TT;
        if (t2==S_INT) return C_OT;
        if (t2==S_LONG) return C_OT;
        return C_NP;
      case S_INT:
        if (t2==S_BYTE) return C_TT;
        if (t2==S_INT) return C_TT;
        if (t2==S_LONG) return C_OT;
        return C_NP;
      case S_FLT:
        if (t2==S_DBL) return C_OT;
        return C_NP;
      case S_DBL:
        if (t2==S_FLT) return C_TT;
        return C_NP;
    }*/
    return C_NP; //no matching standard-conversion found
	}
	
	public char getSig() {
    if (baseType==T_VOID) return 'v';
    if (arrDim!=0) return 'a';
    switch (baseType) {
      case T_BOOL: return 'b';
      case T_BYTE: case T_SHRT: case T_CHAR:
      case T_INT:  return 'i';
      case T_LONG: return 'l';
      case T_FLT:  return 'f';
      case T_DBL:  return 'd';
    }
    return 'o';
  }
	
	public static int getMinSize(int std) {
    switch (std) {
      case T_BYTE: case T_BOOL: return 1;
      case T_SHRT: case T_CHAR: return 2;
      case T_INT:  case T_FLT:  return 4;
      case T_LONG: case T_DBL:  return 8;
    }
    return 0; //this is an error
  }
	
  public boolean isBoolType() {
    return baseType==T_BOOL && arrDim==0;
  }
  
  public boolean isShortType() {
    return baseType==T_SHRT && arrDim==0;
  }
  
  public boolean isIntType() {
    return baseType==T_INT && arrDim==0;
  }
  
  public boolean isLongType() {
    return baseType==T_LONG && arrDim==0;
  }
  
  public boolean isObjType() {
    return baseType==T_NNPT || arrDim!=0 || (baseType!=T_VOID && (qid!=null
        && ((qid.unitDest.modifier&(Unit.M_STRUCT|Unit.M_INDIR))==0)));
  }

  public boolean isIntfType() {
    return baseType==T_NDPT || (arrDim==0 && qid!=null && (qid.unitDest.modifier&Unit.M_INDIR)!=0);
  }
  
  public boolean isStructType() {
    return (qid!=null && (qid.unitDest.modifier&Unit.M_STRUCT)!=0);
  }

  public boolean isThrowableType(Context ctx) {
    return (qid!=null && arrDim==0 && ctx.excThrowable!=null
        && ctx.excThrowable.isParent(qid.unitDest, ctx));
  }
  
  public boolean isCheckedExceptionType(Context ctx) {
    return (qid!=null && arrDim==0 && ctx.excChecked!=null
        && ctx.excChecked.isParent(qid.unitDest, ctx));
  }
  
  public int minimumAccessLevel(String itemMsg, Unit unitContext, Unit destUnit, int destLevel, Context ctx) {
    if (itemMsg==null) itemMsg="item";
    if (unitContext!=destUnit && !unitContext.isOuter(destUnit)
        && !destUnit.isOuter(unitContext)) { //check modifier as call is not inside class
      if (destUnit==null || destUnit.pack==null) {
        printPos(ctx, itemMsg);
        ctx.out.print(" is invalid");
        return L_NOTVISIBLE;
      }
      if (unitContext.pack.packDest!=destUnit.pack.packDest
          && !destUnit.isParent(unitContext, ctx)) { //check modifier to be public
        if ((destLevel&Modifier.M_PUB)==0) {
          printPos(ctx, itemMsg);
          ctx.out.print(ERR_NOTVIS);
          return L_NOTVISIBLE;
        }
        return L_PUBLIC;
      }
      else if (unitContext.pack.packDest!=destUnit.pack.packDest
          && destUnit.isParent(unitContext, ctx)) { //check modifier to be protected
        if ((destLevel&(Modifier.M_PUB|Modifier.M_PROT))==0) { //check modifier to be public or protected
          printPos(ctx, itemMsg);
          ctx.out.print(ERR_NOTVIS);
          return L_NOTVISIBLE;
        }
        return L_PROTECTED;
      }
      else {
        if ((destLevel&(Modifier.M_PUB|Modifier.M_PROT|Modifier.M_PACP))==0) { //check modifier to be public or protected or package private
          printPos(ctx, itemMsg);
          ctx.out.print(ERR_NOTVIS);
          return L_NOTVISIBLE;
        }
        return L_PACKPRIV;
      }
    }
    return L_PRIVATE;
  }
}
