/* Copyright (C) 2005, 2006, 2007, 2008, 2009 Stefan Frenz
 * 
 * This file is part of SJC, the Small Java Compiler written by Stefan Frenz.
 * 
 * SJC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * SJC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with SJC. If not, see <http://www.gnu.org/licenses/>.
 */

package frontend.sjava;

import compbase.*;
import backend.Architecture;
import backend.Instruction;

/**
 * StSwitch: switch-case-statement
 * 
 * @author S. Frenz
 * @version 091005 adopted changed Expr
 *  version 090809 adopted changed Expr, added support for detailed flow analysis
 *  version 090718 adopted changed Expr
 *  version 090508 adopted changes in Stmt
 *  version 090506 added flow analysis information
 *  version 090207 added copyright notice
 *  version 081021 adopted changed Architecture.genComp/genCondJmp
 *  version 080604 better internal error message
 *  version 080525 adopted changed genCondJmp signature
 *  version 080508 added flow hints
 *  version 071108 fixed check for exact type match for constant case-expression
 *  version 070913 added support for spread local variable declaration
 *  version 070909 optimized signature of Expr.resolve, adopted changes in StBreakable
 *  version 070908 optimized signature of Stmt.resolve
 *  version 070809 adopted changed for float/double
 *  version 070628 adopted changes for "Expr extends TypeRef"
 *  version 070625 using StdTypes instead of TypeRef for standard types
 *  version 070114 reduced access level where possible
 *  version 070111 adapted change in printPos and compErr
 *  version 061231 optimized genOutput
 *  version 061229 removed access to firstInstr
 *  version 061203 optimized calls to printPos and compErr
 *  version 060620 added kill-on-jump
 *  version 060607 initial version
 */

public class StSwitch extends StBreakable {
	protected Expr cond;
	protected Stmt stmts, def;
	protected CondStmt caseConds;
	private int constType;
	
	protected StSwitch(StBreakable io, StringList ila, int fid, int il, int ic) {
		super(io, ila, fid, il, ic);
	}
	
	protected int innerResolve(int flowCode, Unit unitContext, Mthd mthdContext, Context ctx) {
	  Stmt stmt;
	  CondStmt cases;
    Vrbl oldVarState;
    int singleRes=FA_NO_FLOWCHANGE, complRes=FA_NO_FLOWCHANGE;
    
    //TODO checking of variable writing is not yet implemented correctly
    
    //remember variable state
    oldVarState=mthdContext.vars;
    //resolve condition
	  if (!cond.resolve(unitContext, mthdContext, true, (flowCode&FA_INSIDE_LOOP)!=0, null, ctx)) {
	    ctx.out.print(" in switch-condition");
	    return FA_ERROR;
	  }
    if (cond.arrDim>0) {
      printPos(ctx, "switch needs basic type");
      return FA_ERROR;
    }
	  switch (cond.baseType) {
	    case StdTypes.T_BYTE:
	    case StdTypes.T_SHRT:
	    case StdTypes.T_CHAR:
	    case StdTypes.T_INT:
	      constType=StdTypes.T_INT;
	      break;
	    case StdTypes.T_LONG:
	      constType=StdTypes.T_LONG;
	      break;
	    default:
	      printPos(ctx, "type of condition must be byte, short, char, int or long");
	      return FA_ERROR;
	  }
	  //resolve cases and statements
	  stmt=stmts;
    cases=caseConds;
	  while (stmt!=null) {
	    //handle case conditions if this is their place
      while (cases!=null && cases.stmt==stmt) {
        //merge current state into complRes
        complRes|=(singleRes&FA_ERROR)|(flowAnalysisBuffer&FA_HAS_ENDBLOCK);
        //reset buffer and singleRes as a new case-block starts
        singleRes=flowAnalysisBuffer=flowCode;
        //resolve expression
        if (!cases.cond.resolve(unitContext, mthdContext, true, (flowCode&FA_INSIDE_LOOP)!=0, cond, ctx)) {
          ctx.out.print(" in case-condition");
          return FA_ERROR;
        }
        if (cases.cond.calcConstantType(ctx)!=constType || cases.cond.baseType!=cond.baseType) {
          printPos(ctx, "condition is not constant or has wrong type (must match switch expression)");
          return FA_ERROR;
        }
        if (cases.stmt==null) {
          printPos(ctx, "null-stmt ");
          return FA_ERROR;
        }
        //next one
        cases=cases.nextCondStmt;
      }
      //handle default entry if this is its place
      if (def==stmt) {
        //merge current state into complRes
        complRes|=(singleRes&FA_ERROR)|(flowAnalysisBuffer&FA_HAS_ENDBLOCK);
        //reset buffer and singleRes as the default-block starts
        singleRes=flowAnalysisBuffer=flowCode;
      }
      //handle statement
	    if (((singleRes=stmt.resolve(singleRes, unitContext, mthdContext, ctx))&FA_ERROR)!=0) {
	      ctx.out.print(" in switch-statement");
	      return FA_ERROR;
	    }
	    stmt=stmt.nextStmt;
	  }
    //last block was potentially not merged into complRes
    complRes|=singleRes&(FA_HAS_ENDBLOCK|FA_ERROR);
    if ((complRes&FA_HAS_ENDBLOCK)==0 && (singleRes&FA_NEXT_IS_UNREACHABLE)!=0 && def!=null) flowCode|=FA_NEXT_IS_UNREACHABLE;
    if ((complRes&FA_ERROR)!=0) flowCode|=FA_ERROR;
    //restore variable state
	  mthdContext.vars=oldVarState;
    return flowCode;
	}
	
	protected void genOutput(Context ctx) {
	  Stmt stmt;
	  CondStmt condstmt;
	  int reg, id, condHnd;
    Instruction defIns=null;
	  
	  breakDest=ctx.arch.getUnlinkedInstruction();
	  id=ctx.arch.insertFlowHint(Architecture.F_BLOCKSTART, 0);
	  ctx.arch.insertFlowHint(Architecture.F_CONDSTART, id);
	  if (ctx.arch.prepareFreeReg(0, 0, 0, cond.baseType)!=0) { //should be 0 always
	    compErr(ctx, "no free reg at beginning of switch");
	    return;
	  }
	  reg=ctx.arch.allocReg();
	  cond.genOutputVal(reg, ctx);
    ctx.arch.insertKillHint(reg);
	  condstmt=caseConds;
	  while (condstmt!=null) {
	    if (constType==StdTypes.T_INT) condHnd=ctx.arch.genCompValToConstVal(reg,
	        condstmt.cond.getConstIntValue(ctx), cond.baseType, Ops.C_EQ);
	    else condHnd=ctx.arch.genCompValToConstDoubleOrLongVal(reg,
	        condstmt.cond.getConstLongValue(ctx), false, Ops.C_EQ);
	    ctx.arch.genCondJmp(condstmt.stIns=ctx.arch.getUnlinkedInstruction(), condHnd);
	    condstmt=condstmt.nextCondStmt;
	  }
	  ctx.arch.deallocRestoreReg(reg, 0, 0);
	  if (def!=null) {
      defIns=ctx.arch.getUnlinkedInstruction();
      ctx.arch.genJmp(defIns);
    }
	  else ctx.arch.genJmp(breakDest);
	  stmt=stmts;
    condstmt=caseConds;
    ctx.arch.insertFlowHint(Architecture.F_SWSBSTART, id);
	  while (stmt!=null) {
      while (condstmt!=null && condstmt.stmt==stmt) {
        ctx.arch.appendInstruction(condstmt.stIns);
        condstmt.stIns=null;
        condstmt=condstmt.nextCondStmt;
      }
      if (def==stmt) ctx.arch.appendInstruction(defIns);
	    stmt.genOutput(ctx);
	    stmt=stmt.nextStmt;
	  }
	  ctx.arch.appendInstruction(breakDest);
    ctx.arch.insertFlowHint(Architecture.F_BLOCKEND, id);
    breakDest=null;
	}
}
